/**** SimCore since 2002-05-15 by Kenji KISE ****/ 
/************************************************/
#include "scalar.h"

/************************************************/
#define WEAKLY_TAKEN 2  /* Initial Value */
Bpred::Bpred(Env_s *ev){
  size = ev->sc2->btb_entry;
  ent = new class btb_entry[size];
  for(int i=0; i<size; i++){
    ent[i].tag     = 0;
    ent[i].target  = 0;
    ent[i].predict = WEAKLY_TAKEN;
  }
}

/************************************************/
void Bpred::Init(chip_wire *w){}

/************************************************/
void Bpred::Step(chip_wire *w){
  ADDR_TYPE pc = w->pcgen_pc;
  int index = (pc>>2) % size;
  btb_entry *e = &ent[index];
  int bc = e->predict;
  w->bpred_tkn = (e->tag==pc && bc>=2) ? 1 : 0;
  w->bpred_targ = e->target;
}

/************************************************/
void Bpred::Update(chip_wire *w){
  if(w->bpred_wpc!=0){
    int taken     = w->bpred_wtkn;
    ADDR_TYPE pc  = w->bpred_wpc;
    ADDR_TYPE npc = w->bpred_wnpc;
    int index = (pc>>2) % size;
    btb_entry *e = &ent[index];
    
    if(e->tag==pc){
      if( taken && e->predict<3) e->predict++;
      if(!taken && e->predict>0) e->predict--;
    }
    else{
      e->tag     = pc;
      e->predict = WEAKLY_TAKEN;
    }
    
    e->target = (taken) ? npc : e->target;
  }
}

/** PC Generation Unit                         **/
/************************************************/
PCgen::PCgen(Env_s *e){ /* Constructor */
  ev = e;
  pc= e->as->pc; /* Entry PC */
}

/************************************************/
void PCgen::Init(chip_wire *w){}

/************************************************/
void PCgen::Step(chip_wire *w){
  w->pcgen_npc = pc+4;
}

/************************************************/
void PCgen::Update(chip_wire *w){
  int stall =  w->data_stall | w->code_stall |
    w->icache_stall | w->dcache_stall;

  pc = 
    (w->pcgen_pal) ? w->pcgen_pal       : 
    (w->pcgen_bpr) ? w->pcgen_bpr       :
    (stall)             ? pc            :
    (w->bpred_tkn)      ? w->bpred_targ : 
    w->pcgen_npc;
}
/************************************************/
